import copy
import operator
import os
import numpy as np
import wandb
import sys
from importlib import import_module
import keras
import keras.backend as K


class WandbKerasCallback(keras.callbacks.Callback):
    """WandB Keras Callback.

    Automatically saves wandb-history.csv and wandb-summary.csv, both tracking
    keras metrics.

    Optionally saves the best model while training.

    Optionally logs weights and gradients during training.

    """

    def __init__(self, monitor='val_loss', verbose=0, mode='auto',
                 save_weights_only=False, log_weights=False, log_gradients=False,
                 save_model=True, training_data=None
                 ):
        """Constructor.

        # Arguments
            monitor: quantity to monitor.
            mode: one of {auto, min, max}.
                'min' - save model when monitor is minimized
                'max' - save model when monitor is maximized
                'auto' - try to guess when to save the model
            save_weights_only: if True, then only the model's weights will be
                saved (`model.save_weights(filepath)`), else the full model
                is saved (`model.save(filepath)`).
            save_model:
                True - save a model when monitor beats all previous epochs
                False - don't save models
            log_weights: if True save the weights in wandb.history
            log_gradients: if True log the training gradients in wandb.history
            training_data: tuple (X,y) needed for calculating gradients


        """
        if wandb.run is None:
            raise wandb.Error(
                'You must call wandb.init() before WandbKerasCallback()')
        self.validation_data = None

        self.monitor = monitor
        self.verbose = verbose
        self.save_weights_only = save_weights_only

        self.filepath = os.path.join(wandb.run.dir, 'model-best.h5')
        self.save_model = save_model
        self.log_weights = log_weights
        self.log_gradients = log_gradients
        self.training_data = training_data

        if self.training_data:
            if len(self.training_data) != 2:
                raise ValueError("training data must be a tuple of length two")

        # From Keras
        if mode not in ['auto', 'min', 'max']:
            print('WandbKerasCallback mode %s is unknown, '
                  'fallback to auto mode.' % (mode))
            mode = 'auto'

        if mode == 'min':
            self.monitor_op = operator.lt
            self.best = float('inf')
        elif mode == 'max':
            self.monitor_op = operator.gt
            self.best = float('-inf')
        else:
            if 'acc' in self.monitor or self.monitor.startswith('fmeasure'):
                self.monitor_op = operator.gt
                self.best = float('-inf')
            else:
                self.monitor_op = operator.lt
                self.best = float('inf')

    def set_params(self, params):
        self.params = params
        print(params)

    def set_model(self, model):
        self.model = model

    def on_epoch_begin(self, epoch, logs=None):
        pass

    def on_epoch_end(self, epoch, logs=None):
        # history
        row = {'epoch': epoch}
        row.update(logs)

        if self.log_weights:
            weights_metrics = self._log_weights()
            row.update(weights_metrics)

        if self.log_gradients:
            gradients_metrics = self._log_gradients()
            row.update(gradients_metrics)

        wandb.run.history.add(row)

        # summary
        self.current = logs.get(self.monitor)
        if self.current is None:    # validation data wasn't set
            #            print('Can save best model only with %s available, '
            #                  'skipping.' % (self.monitor))
            wandb.run.summary.update(row)
            return

        copied = copy.copy(row)
        if self.monitor_op(self.current, self.best):
            copied.pop('epoch')
            wandb.run.summary.update(copied)
            if self.save_model:
                self._save_model()

    def on_batch_begin(self, batch, logs=None):
        pass

    def on_batch_end(self, batch, logs=None):
        pass

    def on_train_begin(self, logs=None):
        pass

    def on_train_end(self, logs=None):
        pass

    def _log_weights(self):
        metrics = {}
        for layer in self.model.layers:
            weights = layer.get_weights()
            if len(weights) == 1:
                metrics[layer.name] = np.mean(weights[0])
            elif len(weights) == 2:
                metrics[layer.name + ":weights-mean"] = np.mean(weights[0])
                metrics[layer.name + ":bias-mean"] = np.mean(weights[1])
        return metrics

    def _log_gradients(self):
        if (not self.training_data):
            raise ValueError(
                "Need to pass in training data if logging gradients")

        X_train = self.training_data[0]
        y_train = self.training_data[1]
        metrics = {}
        weights = self.model.trainable_weights  # weight tensors
        # filter down weights tensors to only ones which are trainable
        weights = [weight for weight in weights
                   if self.model.get_layer(weight.name.split('/')[0]).trainable]

        gradients = self.model.optimizer.get_gradients(
            self.model.total_loss, weights)  # gradient tensors
        input_tensors = [self.model.inputs[0],  # input data
                         # how much to weight each sample by
                         self.model.sample_weights[0],
                         self.model.targets[0],  # labels
                         K.learning_phase(),  # train or test mode
                         ]

        get_gradients = K.function(inputs=input_tensors, outputs=gradients)

        grads = get_gradients([X_train, np.ones(len(y_train)), y_train])

        for (weight, grad) in zip(weights, grads):
            metrics[weight.name.split(':')[0] + ":grad-mean"] = np.mean(grad)
            metrics[weight.name.split(':')[0] + ":grad-stddev"] = np.std(grad)
        return metrics

    def _save_model(self):
        if self.verbose > 0:
            print('Epoch %05d: %s improved from %0.5f to %0.5f,'
                  ' saving model to %s'
                  % (epoch, self.monitor, self.best,
                     self.current, self.filepath))
        self.best = self.current

        try:
            if self.save_weights_only:
                self.model.save_weights(self.filepath, overwrite=True)
            else:
                self.model.save(self.filepath, overwrite=True)
        except ImportError:
            print("Warning: Can't save model without h5py installed")
            self.save_model = False



import pusherclient
import time
import json
import sys
import click
import requests
import logging
from wandb.api import Api
import wandb

api = Api()
if api.settings('base_url').endswith("dev"):
    key = 'c48d148df77404b3c278'
else:
    key = 'e670693ad14e2af5f1dd'


@staticmethod
def _generate_private_key(socket_id, key, channel_name, secret):
    client = requests.Session()
    client.auth = ('api', api.api_key)
    client.timeout = 5
    client.headers.update({
        'User-Agent': api.user_agent,
    })
    #TODO: retry
    res = client.post(api.settings('base_url') + "/pusher/auth", data={
        "socket_id": socket_id,
        "channel_name": channel_name, "entity": secret
    })
    res.raise_for_status()
    return res.json()["auth"]


# Monkey patch the client
pusherclient.Pusher._generate_private_key = _generate_private_key
pusherclient.Pusher.host = "ws-us2.pusher.com"


class AgentPuller(object):
    def __init__(self, agent, entity=None, logger=None):
        self.agent = agent
        self.entity = entity or api.settings('entity')
        self.logger = logger or logging.getLogger(__name__)
        self.last_event = None
        self.socket_id = None
        self.subscribed = False
        self.subscribing = False
        self.channel_name = "private-agent-" + self.entity
        self.client = pusherclient.Pusher(key, secret=self.entity)
        self.client.connection.bind(
            'pusher:connection_established', self._connect)

    def run(self):
        self.client.connect()
        while True:
            self.subscribed = self.subscribe()
            time.sleep(1)

    def subscribe(self):
        if self.subscribed:
            return True
        if self.socket_id and not self.subscribing:
            try:
                self.subscribing = True
                events = self.client.subscribe(self.channel_name)
                events.bind("event", self.event)
                return True
            finally:
                self.subscribing = False
        return False

    def event(self, data):
        self.last_event = time.time()
        self.logger.debug("received event")
        event = json.loads(data)
        self.agent.handle_event(event)

    # We can't subscribe in here because it swallows all exception,
    # so we subscribe in the main thread
    def _connect(self, data):
        self.logger.info(
            "Connecting to entity {0} for events".format(self.entity))
        self.socket_id = json.loads(data)["socket_id"]


class LogPuller(object):
    def __init__(self, run_id, pod_id=None, finished=lambda: {exit()}):
        self.run_id = run_id
        self.pod_id = pod_id
        self.client = pusherclient.Pusher(key)
        self.started = False
        self.last_message = None
        self.finished = finished
        self.client.connection.bind(
            'pusher:connection_established', self._connect)

    def sync(self):
        self.client.connect()
        times = 0
        while True:
            time.sleep(1)
            times += 1
            timeout = 15 if self.last_message else 30
            if self.last_message and time.time() - self.last_message > timeout:
                try:
                    pod = requests.get(
                        "http://kubed.endpoints.playground-111.cloud.goog/pods/%s" % self.pod_id).json()
                    if not pod["status"]["phase"] in ["Error", "Completed"] and times < 3600:
                        if times % 30 == 0:
                            wandb.termlog("No output from process.  Current status %s, retrying in 30 seconds..." %
                                          pod["status"]["phase"])
                    else:
                        break
                except:
                    logging.error(sys.exc_info())
            elif not self.last_message and times >= 60:
                wandb.termlog(
                    "No logs available \U0001F612 try reconnecting with `wandb logs %s`" % self.run_id)
                break

    def lines(self, data):
        self.last_message = time.time()
        for line in json.loads(data):
            l = line["line"].split(" ")
            l.pop(0)
            sys.stdout.write(" ".join(l))
        sys.stdout.flush()

    def _connect(self, data):
        logs = self.client.subscribe("logs-" + self.run_id)
        logs.bind("lines", self.lines)



#!/usr/bin/env python

from __future__ import print_function

import collections
import contextlib
import copy
import json
import os
import time
from threading import Lock
import warnings
import weakref
import six

torch = None  # dynamically imported by History.torch
import wandb
from wandb import util
from wandb import media


class History(object):
    """Used to store data that changes over time during runs. """

    def __init__(self, fname, out_dir='.', add_callback=None, stream_name="default"):
        self._start_time = wandb.START_TIME
        self.out_dir = out_dir
        self.fname = os.path.join(out_dir, fname)
        self.rows = []
        self.row = {}
        self.stream_name = stream_name
        # during a batched context logging may still be disabled. we do it this way
        # so people don't have to litter their code with conditionals
        self.compute = False
        self.batched = False
        # not all rows have the same keys. this is the union of them all.
        self._keys = set()
        self._streams = {}
        self._steps = 0
        self._lock = Lock()
        try:
            with open(self.fname) as f:
                for line in f:
                    self._index(json.loads(line))
        except IOError:
            pass

        self._file = open(self.fname, 'w')
        self._add_callback = add_callback

    def keys(self):
        return list(self._keys)

    def stream(self, name):
        """stream can be used to record different time series:

        run.history.stream("batch").add({"gradients": 1})
        """
        if self.stream_name != "default":
            raise ValueError("Nested streams aren't supported")
        if self._streams.get(name) == None:
            self._streams[name] = History(self.fname, out_dir=self.out_dir,
                                          add_callback=self._add_callback, stream_name=name)
        return self._streams[name]

    def column(self, key):
        """Fetches a key from all rows that have it. Skips those that don't.
        """
        for row in self.rows:
            if key in row:
                yield row[key]

    def add(self, row, write="auto"):
        """Adds keys to history.  If write is False enables incremental
        history updates:

        run.history.add({"loss": 1}, write=False)
        run.history.row["duration"] = 1.0
        run.history.write()

        If write is True, writes immediatly.  If write is auto, writes if we aren't within
        a step context
        """
        if not isinstance(row, collections.Mapping):
            raise wandb.Error('history.add expects dict-like object')
        # Write the last row if step isn't specified or step is different than what was commited last
        auto_write = write == "auto" and not self.batched
        if auto_write:
            self.write()
        if self.stream_name != "default":
            row["_stream"] = self.stream_name
        self.row.update(row)
        if write == True or auto_write:
            self.write()

    @contextlib.contextmanager
    def step(self, compute=True, stream=None):
        """Context manager to gradually build a history row, then commit it at the end.

        To reduce the number of conditionals needed, code can check run.history.compute:

        run.history.step(batch_idx % log_interval == 0):
            run.history.add({"nice": "ok"})
            if run.history.compute:
                # Something expensive here
        """
        if stream:
            history = self.stream(stream)
        else:
            history = self
        history.row = {}
        history.batched = True
        history.compute = compute
        yield(history)
        if compute:
            history.add(self.row, True)

    def _index(self, row):
        """Internal row adding method that updates step, streams and keys"""
        # TODO: store a downsampled representation in memory
        self.rows.append(row)
        self._keys.update(row.keys())
        self._steps += 1

    def _transform(self):
        """Transforms rich classes into the proper format before writing"""
        for key, val in six.iteritems(self.row):
            if type(val) in (list, tuple):
                if len(val) > 0 and type(val[0]) == media.Image:
                    self.row[key] = media.Image.transform(val, self.out_dir,
                                                          "{}_{}.jpg".format(key, self.row["_step"]))

    def write(self):
        if self.row:
            self._lock.acquire()
            try:
                self.row['_runtime'] = round(time.time() - self._start_time, 2)
                self.row['_step'] = self._steps
                self._transform()
                self._file.write(util.json_dumps_safer(self.row))
                self._file.write('\n')
                self._file.flush()
                self._index(self.row)
                if self._add_callback:
                    self._add_callback(self.row)
                self.row = {}
            finally:
                self._lock.release()
                return True
        else:
            return False

    def close(self):
        self.write()
        self._lock.acquire()
        try:
            self._file.close()
            self._file = None
        finally:
            self._lock.release()
